from .sendfiles import hyperpredictions
from .sendfiles import rundemo
from .sendfiles import algodescription
from .sendfiles import hypertraining
from .sendfiles import abort
from .sendfiles import finddistribution
from .sendfiles import readfile
from .sendfiles import distattr

name = "maadsbml"
